/* open proxy command-line checker.
 * Copyright by Michael Tokarev <mjt@corpit.ru>. Nov-2002.
 * This program is free software.  It may be used and distributed under
 * the terms of Gnu General Public License (GPL) version 2 or later.
 *
 * ToDo:
 *  - a mode in which proxycheck will stop scanning an IP after
 *    first open proxy found on that IP
 *  - wingate "ftp proxy" checking
 *  - GET/PUT/POST HTTP proxy detection (only with a help of ext. prog?)
 *  - telnet proxy (type "telnet host port")
 *  - cisco default password "telnet proxy" detection
 *  - more homework before calling external helper: i.e.
 *    check HTTP reply code, look at "Password:" prompt for wingate
 *    proxy etc.
 *
 * Changes:
 * 10-Dec-2002
 *  - do not use setenv() (not all systems have it) - with more cleanups... ;)
 *  - some more cleanups, better "number of bytes read" calculation
 *  - added -C option (similar to -c but also send a proxy ident to
 *    remote system
 *  - added a program, proxylogger, to be used for -C proxycheck's option
 *  - added another commonly (ab)used http port 8081
 *  - rearranged a list of protocols to be less aggressive
 *
 * 23-Nov-2002
 *  - *many* cleanups, little improvements, ...
 *  - bugfix with not setting FD_CLOEXEC flag properly
 *  - connect both stdin and stdout to remote when executing
 *    external helper - it's a pain to program both Input and Output
 *    on stdin in perl.  Stdout isn't used anyway.
 *  - count number of bytes received when using external helper (it
 *    should write 4-byte number to filedescriptor #3 - a pipe read
 *    by proxycheck)
 * 11-Nov-2002
 *  - added -e "host:port:program" option: similar to -c, but
 *    executes a given program when connection (seems to be)
 *    established, with fd 0 connected to proxy, and with
 *    several environment variables set.  If this program
 *    will return 0, proxycheck assumes the test was successeful.
 *  - added dsblsmtp program which can be used with newly added
 *    -e option to automatically submit any proxies found to the
 *    DSBL.org blocklist.  This is in early beta, but works for
 *    me.  It expects two environment variables, DSBL_USER and
 *    DSBL_PASS, to be set in order to autorize access to dsbl.
 *    In order to use it, connect to mx.listme.dsbl.org:25.
 *  - wingate3.0 default telnet proxy is at port 1181, not 1183,
 *    and ftp proxy is at port 1183.  My mistake.
 *  - miscellaneous cleanups/corrections/bugfixes
 *
 * 06-Nov-2002
 *  - rearranged protocol handlers to send commands to proxies
 *    one-by-one (e.g. socks5 auth and conn requests), and to
 *    read and interpret proxy replies (e.g. socks status codes).
 *    This change allows detection of socks5 proxy servers who
 *    ignores all but first 3 bytes in first tcp packet sent
 *    (before we sent two commands in one packet)
 *  - added a check for new filedescriptors to not be greather
 *    than FD_SETSIZE, since event library can't handle that.
 *  - misc. cleanups
 *
 * 31-Oct-2002
 *  - Added a temporary workaround to detect HwaRang http proxies
 *  - removed usage of inet_aton() in flavour of inet_addr() in order
 *    to not require -lresolv on Solaris
 *  - added new ports to the lists (wingate 3.0):
 *    1180 - socks
 *    1182 - http
 *    1183 - wingate
 *   Yet to be done is wingate's ftp proxy on 1181 - but for this,
 *   complete rewrite is necessary (it is a very special proxy...)
 *
 *  03-Oct-2002
 *  - Corrected bug with high port numbers (proxycheck was unable to
 *    use ports > 32767, Thanks to Matt Christian @visi.com.
 */

#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/wait.h>
#include "event.h"

#ifndef __GNUC__
# define __attribute__(x)
#endif
#define UNUSED __attribute__((unused))

#ifndef INADDR_NONE
# define INADDR_NONE ((in_addr_t)-1)
#endif

typedef struct proxyhost proxyhost_t;
typedef struct proxyport proxyport_t;
typedef struct proxyconn proxyconn_t;
typedef struct proxyproto proxyproto_t;
typedef struct proxyproc proxyproc_t;

struct proxyport {
  unsigned short port;	/* host byte order! */
  const proxyproto_t *proto;
};
#define proxynoport(p) ((p)->port == 0)

struct proxyhost {
  struct in_addr addr;
  int nopen;
};
#define proxynohost(h) ((h)->addr.s_addr == 0)

typedef struct proxydest { /* a destination request for a proxy */
  struct in_addr addr;	/* address to connect to */
  unsigned short port;	/* destination port, host byte order */
  int type;
#define DT_EXPECT  0 /* expect a string (-c) */
#define DT_TEXPECT 1 /* send tag, expect a string (-C) */
#define DT_EXEC    2 /* run a program (-e) */
  char *str;		/* expect string or program to execute */
  int slen;		/* length of str */
} proxydest_t;
#define proxynodest(d) ((d)->port == 0)

typedef struct proxybuf {
  int len;
  char buf[128 - sizeof(int)];
} proxybuf_t;

struct proxyconn {
  proxyhost_t *host;
  proxyport_t *port;
  proxydest_t *dest;
  proxybuf_t *buf;
  int nread;
  int state;
};

struct proxyproc {
  pid_t pid;
  proxyproc_t *next;
  proxyhost_t *host;
  proxyport_t *port;
  proxydest_t *dest;
};

typedef union proxybuflist {
  proxybuf_t buf;
  union proxybuflist *next;
} proxybuflist_t;    

typedef union proxyconnlist {
  proxyconn_t conn;
  union proxyconnlist *next;
} proxyconnlist_t;

typedef struct {
  proxyhost_t *hosts, *chost; /* array of hosts to check */
  proxyport_t *ports, *cport; /* array of ports to check */
  proxydest_t *dests, *cdest; /* array of destinations to try */
  proxyconnlist_t *conns; /* poll of space conn structs */
  proxybuflist_t *bufs; /* poll of space buf structs */
  proxyproc_t *procs; /* list of currently running processes */
  int spipe[2]; /* child's status pipe */
  int ntodo; /* number of todo conns */
  int ndone; /* number of completed checks */
  int nactive; /* number of active connections */
  int nopen; /* number of open IPs found */
  long nread; /* total number of bytes read */
} proxystate_t;

struct proxyproto {
  const char *name;
  int type;
  const char *defports;
};

char *progname;
int verbose;
struct sockaddr_in bindaddr;
int timeout = 60;
int maxconn;

static proxystate_t _state;
static proxystate_t *state = &_state;

static volatile int child_done;
static void sigchild(int sig) {
  child_done = 1;
#ifndef SA_RESTART
  signal(SIGCHLD, sigchild);
#endif
}

static void reapchildren(proxystate_t *state, int opt);

void __attribute__((noreturn,format(printf,2,3)))
error(int errnum, const char *fmt, ...) {
  char buf[128];
  int l = sprintf(buf, "%s: ", progname);
  va_list ap;
  va_start(ap, fmt); l += vsprintf(buf + l, fmt, ap); va_end(ap);
  if (errnum)
    l += sprintf(buf + l, ": %s\n", strerror(errnum));
  else
    buf[l++] = '\n';
  write(2, buf, l);
  exit(1);
}

void *xrealloc(void *ptr, size_t sz) {
  ptr = ptr ? realloc(ptr, sz) : malloc(sz);
  if (!ptr)
    error(0, "unable to allocate %d bytes of memory", sz);
  return ptr;
}
#define xmalloc(sz) xrealloc(NULL, sz)

static void __attribute__((format(printf,3,4)))
cinfo(int level, const proxyconn_t *conn, const char *fmt, ...) {
  if (level <= verbose) {
    va_list ap;
    char buf[128];
    int l;
    if (conn)
      l = sprintf(buf, "%s:%s:%d: ", conn->port->proto->name,
                  inet_ntoa(conn->host->addr), conn->port->port);
    else
      l = 0;
    va_start(ap, fmt); l += vsprintf(buf + l, fmt, ap); va_end(ap);
    buf[l++] = '\n';
    write(2, buf, l);
  }
}

static void readspipe(event_ctx *ec, int fd, event_t event, void *c) {
  int sz;
  if (read(fd, &sz, sizeof(sz)) != sizeof(sz)) {
    event_request_io(ec, fd, 0, NULL, NULL);
    close(fd);
    state->spipe[1] = 0;
  }
  else
    state->nread += sz;
}

static struct in_addr hostaddr(const char *host) {
  struct in_addr iaddr;
  if ((iaddr.s_addr = inet_addr(host)) == INADDR_NONE) {
    struct hostent *he = gethostbyname(host);
    if (!he)
      error(0, "unknown host(name) `%s'", host);
    if (he->h_addrtype != AF_INET)
      error(0, "not INET address %s", host);
    if (he->h_length != sizeof(iaddr))
      error(0, "unexpected length of address of %s", host);
    memcpy(&iaddr, he->h_addr_list[0], sizeof(iaddr));
  }
  return iaddr;
}

static unsigned short portnum(const char *port) {
  int n = atoi(port);
  if (n <= 0 || n > 65535) {
    struct servent *se = getservbyname(port, "tcp");
    if (!se)
      error(0, "invalid port %s", port);
    return ntohs(se->s_port);
  }
  else
    return (unsigned short)n;
}

static void
addports(proxyport_t **portsp, int *nportsp,
         const proxyproto_t *proto, const char *port) {
  const char *space = ",;/ \t";
  int nports = *nportsp;
  proxyport_t *ports = *portsp;
  if (!port)
    port = proto->defports;
  port = strcpy(alloca(strlen(port) + 1), port);
  for(port = strtok((char*)port, space); port; port = strtok(NULL, space)) {
    unsigned short n = portnum(port);
    int i;
    for (i = nports - 1; i >= 0; --i)
      if (ports[i].port == n && ports[i].proto == proto)
        break;
    if (i >= 0)
      continue;
    if (!(nports & 7))
      *portsp = ports =
        (proxyport_t*)xrealloc(ports, sizeof(proxyport_t) * (nports + 9));
    ports[nports].port = n;
    ports[nports].proto = proto;
    ++nports;
  }
  *nportsp = nports;
}

static const proxyproto_t protos[] = {
#define PROXY_SOCKS4 0
  { "socks4", PROXY_SOCKS4, "1080,1180" },
#define PROXY_HTTP 1
  { "http", PROXY_HTTP, "80,81,1182,3128,4480,6588,7033,8000,8080,8081,8085,8090,8095,8100,8105,8110" },
#define PROXY_SOCKS5 2
  { "socks5", PROXY_SOCKS5, "1080,1180" },
#define PROXY_WINGATE 3
  { "wingate", PROXY_WINGATE, "23,1181" },
  { NULL, 0, NULL }
};

static void usage() {
  int i;
  printf(
"%s: open proxy checker.  Usage: %s [options] host...\n"
"options:\n"
"  -c host:port:expect - try to connect to this host:port and wait for\n"
"    expect string (may be used more than once)\n"
"  -C host:port:expect - the same as -c but also send a proxy tag\n"
"    (proto:proxyip:proxyport) to remote system - so that destination\n"
"    system will know information about a proxy\n"
"  -e host:port:program-to-run - try to connect to this host:port and\n"
"    execute program-to-run with stdin connected to proxy server\n"
"    (may be used more than once)\n"
"  -p [proto:]port,port,... - try those protocol/ports\n"
"    if proto: part is omitted, all supported protocols will be probed.\n"
"    if port list part is omitted, default ports for the protocol will\n"
"    be probed (this option may be used more than once)\n"
"  -v - increase verbosity level\n"
"  -t timeout - set timeout in seconds (default 60)\n"
"  -b address - bind to this IP address\n"
"  -m maxconn - do not open more than this number of parallel connections\n"
"  -h - print this help and exit\n"
"Either -e or -c option is required.\n"
"Supported protocols and default ports are:\n"
, progname, progname);
  for (i = 0; protos[i].name; ++i)
    printf("  %s (%s)\n", protos[i].name, protos[i].defports);
  exit(0);
}

static int
init(int argc, char **argv, proxystate_t *state) {
  int c, i;
  char *s;
  char *bindhost = NULL;
  proxyport_t *ports;
  proxyhost_t *hosts;
  proxydest_t *dests;
  int nports, ndests, nhosts;

  if ((progname = strrchr(argv[0], '/')) != NULL)
    argv[0] = ++progname;
  else
    progname = argv[0];

  if (argc < 2)
    usage();

  ports = NULL; nports = 0;
  dests = NULL; ndests = 0;

  state->spipe[0] = state->spipe[1] = 0;

  while((c = getopt(argc, argv, "c:C:e:b:t:p:m:vh")) != EOF)
    switch(c) {
    case 'v': ++verbose; break;
    case 'b': bindhost = optarg; break;
    case 't':
      if ((timeout = atoi(optarg)) < 0 || timeout > 1000)
        error(0, "invalid timeout %s", optarg);
      break;
    case 'm':
      if ((maxconn = atoi(optarg)) < 0)
        error(0, "invalid maxconn value %s", optarg);
      break;
    case 'p':
      if ((s = strchr(optarg, ':')) != NULL)
        *s++ = '\0';
      else
        s = optarg, optarg = "";
      if (!*s || strcmp(s, "*") == 0)
        s = NULL;
      if (!optarg || !*optarg || strcmp(optarg, "*") == 0) {
        for(c = 0; protos[c].name; ++c)
          addports(&ports, &nports, &protos[c], s);
      }
      else {
        for(c = 0; strcasecmp(protos[c].name, optarg) != 0;)
          if (!protos[++c].name)
            error(0, "unknown protocol `%s'", optarg);
        addports(&ports, &nports, &protos[c], s);
      }
      break;
    case 'c':  /* host:port:expect */
    case 'C':  /* ditto, but also send a proxy tag to remote */
    case 'e':  /* host:port:program-to-run */
      if ((s = strchr(optarg, ':')) != NULL) {
        char *p, *e;
        *s++ = 0; p = s;
        if ((s = strchr(s, ':')) != NULL) {
           *s++ = 0; e = s;
           if (*optarg && *p && *e) {
             proxydest_t *d = dests = (proxydest_t*)
              xrealloc(dests, (ndests + 2) * sizeof(proxydest_t));
             d += ndests++;
             d->addr = hostaddr(optarg);
             d->port = portnum(p);
             d->str = e;
             d->slen = strlen(e);
             if (c == 'e') {
               state->spipe[0] = 1;
               d->type = DT_EXEC;
             }
             else
               d->type = (c == 'C') ? DT_TEXPECT : DT_EXPECT;
             continue;
           }
        }
      }
      error(0, "invalid connect specification given");
    case 'h': usage();
    default: error(0, "use -h option for help");
    }

  if (!ndests)
    error(0, "no -c/-C/-e option given");
  memset(&dests[ndests], 0, sizeof(*dests));
  state->dests = state->cdest = dests;

  argv += optind; argc -= optind;
  if (!argc)
    usage();

  /* bind address */
  bindaddr.sin_family = AF_INET;
  if (bindhost)
    bindaddr.sin_addr = hostaddr(bindhost);

  /* default list of protos/ports */
  if (!nports)
    for(c = 0; protos[c].name; ++c)
      addports(&ports, &nports, &protos[c], protos[c].defports);
  ports[nports].port = 0;
  ports[nports].proto = 0;
  state->ports = state->cport = ports;

  /* hosts to check */
  hosts = (proxyhost_t*)xmalloc(sizeof(proxyhost_t) * (argc + 1));
  nhosts = 0;
  for(c = 0; c < argc; ++c) {
    struct in_addr addr = hostaddr(argv[c]);
    for (i = nhosts - 1; i >= 0; --i)
      if (hosts[i].addr.s_addr == addr.s_addr)
        break;
    if (i >= 0)
      continue;
    hosts[nhosts].addr = addr;
    hosts[nhosts].nopen = 0;
    ++nhosts;
  }
  hosts[nhosts].addr.s_addr = 0;
  state->hosts = state->chost = hosts;

  state->bufs = NULL;
  state->conns = NULL;
  state->procs = NULL;

  state->ntodo = nhosts * nports * ndests;
  state->ndone = 0;
  state->nactive = 0;
  state->nopen = 0;

  if (state->spipe[0]) {
    if (pipe(state->spipe) < 0 ||
        fcntl(state->spipe[0], F_SETFD, 1) < 0 ||
        (state->spipe[1] != 3 && fcntl(state->spipe[1], F_SETFD, 1) < 0))
      error(errno, "unable to create pipe");
    event_request_io(NULL, state->spipe[0], EVENT_READ,
                     readspipe, NULL);
  }

#ifdef SA_RESTART
  {
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = SIG_IGN;
    sigaction(SIGPIPE, &sa, NULL);
    sa.sa_handler = sigchild;
    sa.sa_flags = SA_RESTART;
    sigaction(SIGCHLD, &sa, NULL);
  }
#else
  signal(SIGPIPE, SIG_IGN);
  signal(SIGCHLD, sigchild);
#endif

  return nhosts;
}

static proxybuf_t *bufalloc() {
  proxybuflist_t *bl = state->bufs;
  if (!bl) {
    int i;
    bl = (proxybuflist_t*)xmalloc(10 * sizeof(proxybuflist_t));
    for (i = 0; i < 9; ++i)
      bl[i].next = &bl[i+1];
    bl[9].next = NULL;
    state->bufs = bl;
  }
  state->bufs = bl->next;
  bl->buf.len = 0;
  return &bl->buf;
}

static void connh(event_ctx *ec, int fd, event_t event, void *ctx);

static void timeouth(event_ctx *ec, int UNUSED fd,
                     event_t UNUSED event, void *ctx) {
  errno = ETIMEDOUT;
  event_io_wakeup(ec, (int)ctx, EVENT_ERROR);
}

static int nextconn(event_ctx *ec) {
  proxyconnlist_t *cl;
  proxyconn_t *c;
  struct sockaddr_in sin;
  int fd;

  fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (fd < 0) {
    if (errno == EMFILE) {
      if (!maxconn || maxconn > state->nactive)
        maxconn = state->nactive;
      return 0;
    }
    error(errno, "unable to create socket");
  }
  if (fd >= FD_SETSIZE) {
    close(fd);
    if (!maxconn || maxconn > state->nactive)
      maxconn = state->nactive;
    return 0;
  }
  if (bindaddr.sin_addr.s_addr &&
      bind(fd, (struct sockaddr *)&bindaddr, sizeof(bindaddr)) < 0)
    error(errno, "unable to bind");
  if (fcntl(fd, F_SETFL, O_NONBLOCK|O_RDWR) < 0 ||
      fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
    error(errno, "unable to set socket options");

  /* allocate new connection struct */
  if ((cl = state->conns) == NULL) {
    int i;
    cl = (proxyconnlist_t*)xmalloc(50 * sizeof(proxyconnlist_t));
    for(i = 0; i < 49; ++i)
      cl[i].next = &cl[i+1];
    cl[49].next = NULL;
    state->conns = cl;
  }
  state->conns = cl->next;
  c = &cl->conn;
  c->host = state->chost;
  c->port = state->cport;
  c->dest = state->cdest;
  c->buf = NULL;
  c->nread = 0;
  c->state = -1;

  ++state->nactive;
  --state->ntodo;
  if (state->ntodo) {
    if (proxynohost(++state->chost)) {
      state->chost = state->hosts;
      if (proxynodest(++state->cdest)) {
        state->cdest = state->dests;
        ++state->cport;
      }
    }
  }

  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(c->port->port);
  sin.sin_addr = c->host->addr;
  if (connect(fd, (struct sockaddr *)&sin, sizeof(sin)) == 0)
    connh(ec, fd, EVENT_WRITE, c);
  else if (errno != EINPROGRESS)
    connh(ec, fd, EVENT_ERROR, c);
  else if (event_request_io(ec, fd, EVENT_WRITE, connh, c) < 0 ||
           event_request_timer(ec, timeout, timeouth,
                               (void*)fd, event_time(ec)) < 0)
    error(errno, "unable to register I/O event");
  return 1;
}

static void connfree(event_ctx *ec, int fd, proxyconn_t *c, int drain) {
  proxyconnlist_t *cl;
  event_request_io(ec, fd, 0, NULL, NULL);
  event_request_timer(ec, -1, timeouth, (void*)fd, 0);
  if (drain) {
    char buf[1024];
    int l;
    while((l = read(fd, buf, sizeof(buf))) > 0)
      c->nread += l;
  }
  close(fd);
  state->nread += c->nread;
  --state->nactive;
  ++state->ndone;
  if (c->buf) {
    proxybuflist_t *bl = (proxybuflist_t*)c->buf;
    bl->next = state->bufs;
    state->bufs = bl;
  }
  cl = (proxyconnlist_t*)c;
  cl->next = state->conns;
  state->conns = cl;
  if (state->ntodo)
    nextconn(ec);
  else if (!state->nactive && state->spipe[1])
    close(state->spipe[1]);
}

int main(int argc, char **argv) {
  time_t start;

  argc = init(argc, argv, state);

  cinfo(1, NULL, "to check: %d hosts/ports/dests combos, %d hosts",
	state->ntodo, argc);

  start = time(NULL);

  while(state->ntodo && (!maxconn || maxconn > state->nactive))
    if (!nextconn(NULL))
      break;

  while(event_events(NULL)) {
    if (event_wait(NULL, -1) < 0 && errno != EINTR)
      error(errno, "event_wait");
    if (child_done)
      reapchildren(state, WNOHANG);
  }
  if (state->procs) {
    signal(SIGCHLD, SIG_DFL);
    reapchildren(state, 0);
  }

  cinfo(1, NULL, "IPs with open proxies found: %d", state->nopen);
  cinfo(1, NULL, "number of bytes read: %ld", state->nread);
  cinfo(1, NULL, "elapsed time: %d sec", time(NULL) - start);
  return 0;
}

static void found(proxyhost_t *host, proxyport_t *port, proxydest_t *dest) {
  printf("%s %s(%d) open\n", inet_ntoa(host->addr),
         port->proto->name, port->port);
  if (!host->nopen++)
    ++state->nopen;
}

static void expecth(event_ctx *ec, int fd, event_t event, void *ctx) {
  proxyconn_t *c = (proxyconn_t*)ctx;
  char buf[1024];
  int l = c->buf ? c->buf->len : 0;
  if (event != EVENT_READ || (l = read(fd, buf + l, sizeof(buf) - 1 - l)) < 0)
    cinfo(3, c, "read target reply: %s", strerror(errno));
  else if (!l)
    cinfo(2, c, "seems not to be open");
  else {
    c->nread += l;
    if (c->buf) {
      memcpy(buf, c->buf->buf, c->buf->len);
      l += c->buf->len;
    }
    if (memmem(buf, l, c->dest->str, c->dest->slen) != 0)
      found(c->host, c->port, c->dest);
    else if (c->nread > 1024)
      cinfo(3, c, "too much input, closing connection");
    else {
      char *p;
      if (!c->buf && c->port->proto->type == PROXY_HTTP &&
          memmem(buf, l, " HwaRang/", 9) != 0) {
        cinfo(1, c, "probably HwaRang HTTP proxy, workaround enabled");
        /*if (c->type != DT_TEXPECT) -- we wrote something but oh well! */
          write(fd, "\r\n", 2);
      }
      if (l >= c->dest->slen) {
        p = buf + l - c->dest->slen + 1;
        l = c->dest->slen - 1;
      }
      else
        p = buf;
      while(l)
        if (memcmp(p, c->dest->str, l) == 0)
          break;
        else
          --l, ++p;
      if (l) {
        if (!c->buf)
          c->buf = bufalloc();
        memcpy(c->buf->buf, p, l);
        c->buf->len = l;
      }
      else if (c->buf)
        c->buf->len = 0;
      return;
    }
  }
  connfree(ec, fd, c, event == EVENT_READ && l > 0);
}

static void reapchildren(proxystate_t *state, int opt) {
  pid_t pid;
  int status;
  while((pid = waitpid(-1, &status, opt)) > 0) {
    proxyproc_t **pp = &state->procs;
    proxyproc_t *p;
    for(;;) {
      if (!(p = *pp)) {
        fprintf(stderr, "%s: warning: pid %d not found\n", progname, pid);
        break;
      }
      else if (p->pid == pid) {
        if (!status)
          found(p->host, p->port, p->dest);
        *pp = p->next;
        free(p);
        break;
      }
      else
        pp = &p->next;
    }
  }
  child_done = 0;
}


static int vse(const char *var, const char *valfmt, ...) {
  va_list ap;
  char buf[128], *e;
  int l = sprintf(buf, "%s=", var);
  va_start(ap, valfmt);
  l += vsprintf(buf + l, valfmt, ap) + 1;
  va_end(ap);
  e = (char*)malloc(l);
  if (!e) return -1;
  memcpy(e, buf, l);
  putenv(e);
  return 0;
}

static void execprog(event_ctx *ec, int fd, proxyconn_t *c, int ok) {
  pid_t pid;
  pid = fork();
  if (pid < 0)
    error(errno, "unable to fork");
  else if (pid) {
    proxyproc_t *p = (proxyproc_t*) xmalloc(sizeof(proxyproc_t));
    p->pid = pid;
    p->host = c->host;
    p->port = c->port;
    p->dest = c->dest;
    p->next = state->procs;
    state->procs = p;
    connfree(ec, fd, c, 0);
  }
  else { /* child */
    if (fcntl(fd, F_SETFL, O_RDWR) < 0)
      error(errno, "unable to set socket options");
    if (dup2(fd, 0) < 0 || dup2(fd, 1) < 0 ||
        (state->spipe[1] != 3 && dup2(state->spipe[1], 3) < 0))
      error(errno, "unable to redirect input/output");
    vse("VERBOSE", "%d", verbose);
    vse("PROXYADDR", "%s", inet_ntoa(c->host->addr));
    vse("PROXYPORT", "%d", c->port->port);
    vse("PROXYPROTO", "%s", c->port->proto->name);
    vse("PROXYTAG", "%s:%s:%d",
       c->port->proto->name, inet_ntoa(c->host->addr), c->port->port);
    vse("DESTADDR", "%s", inet_ntoa(c->dest->addr));
    vse("DESTPORT", "%d", c->dest->port);
    vse("RELIABLE", ok ? "1" : "0");
    vse("TIMEOUT", "%d", timeout);
    cinfo(3, c, "executing %s (pid=%d)", c->dest->str, getpid());
    execlp(c->dest->str, c->dest->str, NULL);
    error(errno, "unable to execute %s", c->dest->str);
  }
}

static void connh(event_ctx *ec, int fd, event_t event, void *ctx) {
  proxyconn_t *c = (proxyconn_t*)ctx;
  if (event != EVENT_ERROR) {
    char buf[128];
#define DO_NONE 0
#define DO_CONN 1
#define DO_DEST 2
#define DO_DESTUN 3
    int l = 0, next = DO_NONE;
/* Be as dumb as possible.  Do not attempt to parse any replies.
   Quite a few proxy "servers" sends back bogus reply codes (if at all)
   but still allows to connect.  Also, most today's spamaware also
   very dumb, so it's better to "emulate" it here anyway */
    switch(c->port->proto->type) {

    case PROXY_HTTP:
      if (c->state < 0) {
        l = sprintf(buf, "CONNECT %s:%d HTTP/1.0\r\n\r\n",
                    inet_ntoa(c->dest->addr), c->dest->port);
        next = DO_CONN;
	/* next = DO_DESTUN; */
      }
      else { /* read first line of response */
        char *p;
	errno = 0;
        l = read(fd, buf, 14); /* HTTP/1.0 200 etc... */
	if (l <= 0) break;
	c->nread += l;
	buf[l] = 0;
	l = 0;
	p = strchr(buf, ' ');
	if (p && *++p == '2') {
          next = DO_DESTUN;
	  cinfo(2, c, "request ok (%.3s)", p);
	}
	else {
          cinfo(2, c, "seems not to be open (%.3s)", p);
	  connfree(ec, fd, c, 1);
	  return;
	}
      }
      break;

    case PROXY_SOCKS4:
      if (c->state < 0) { /* construct socks4 request */
        unsigned short p = htons(c->dest->port);
        buf[0] = 4; /* socks version */
        buf[1] = 1; /* request: connect */
        memcpy(buf+2, &p, 2);
        memcpy(buf+4, &c->dest->addr.s_addr, 4);
        buf[8] = 0; /* empty/null username */
        l = 9;
        next = DO_CONN;
      }
      else { /* read socks4 reply */
        errno = 0;
        if (read(fd, buf, 8) < 8)
          break;
        c->nread = 8;
        switch(buf[1]) {
	case 90: cinfo(2, c, "socks4: request granted"); next = DO_DEST; break;
	case 91: cinfo(2, c, "socks4: request rejected or failed"); break;
	case 92: cinfo(2, c, "socks4: identd required"); break;
	case 93: cinfo(2, c, "socks4: identd info mismatch"); break;
	default:
	  cinfo(2, c, "socks4: unexpected response v=%d c=%d, trying anyway",
		buf[0], buf[1]);
	  next = DO_DESTUN;
	  break;
	}
	if (!next) {
	  connfree(ec, fd, c, 1);
	  return;
	}
      }
      break;

    case PROXY_SOCKS5:
      if (c->state < 0) { /* request socks5 auth */
	buf[0] = 5; /* socks version */
	buf[1] = 1; /* nmethods */
	buf[2] = 0; /* no auth reqd */
	l = 3;
	next = DO_CONN;
      }
      else if (c->state == 0) { /* read auth reply, send conn req */
	unsigned short p;
	errno = 0;
	if (read(fd, buf, 2) < 2)
	  break;
	if (buf[1] != 0)
	  cinfo(2, c, "auth rejected (v=%d c=%d), trying anyway",
                buf[0], buf[1]);
	p = htons(c->dest->port);
	buf[0] = 5; /* version */
	buf[1] = 1; /* connect cmd */
	buf[2] = 0; /* reserved */
	buf[3] = 1; /* ipv4 addrtype */
	memcpy(buf+4, &c->dest->addr.s_addr, 4);
	memcpy(buf+8, &p, 2);
	l = 10;
	next = DO_CONN;
      }
      else /*if (c->state == 1)*/ { /* read conn req reply */
        errno = 0;
	if (read(fd, buf, 2) < 2) /* only first 2 bytes matters */
	  break;
	c->nread = 4;
	switch(buf[1]) {
	case 0: cinfo(2, c, "socks5: request granted"); next = DO_DEST; break;
	case 1: cinfo(2, c, "socks5: general server failure"); break;
	case 2: cinfo(2, c, "socks5: connection not allowed by ruleset"); break;
	case 3: cinfo(2, c, "socks5: network unreachable"); break;
	case 4: cinfo(2, c, "socks5: host unreachable"); break;
	case 5: cinfo(2, c, "socks5: connection refused"); break;
	case 6: cinfo(2, c, "socks5: TTL expired"); break;
	case 7: cinfo(2, c, "socks5: command not supported"); break;
	case 8: cinfo(2, c, "socks5: address type not supported"); break;
	default:
	  cinfo(2, c, "socks5: unexpected response: v=%d c=%d, trying anyway",
		buf[0], buf[1]);
	  next = DO_DESTUN;
	  break;
	}
	if (!next) {
	  connfree(ec, fd, c, 1);
	  return;
	}
      }
      break;

    case PROXY_WINGATE:
      l = sprintf(buf, "%s:%d\r\n", inet_ntoa(c->dest->addr), c->dest->port);
      next = DO_DESTUN;
      break;

    default: /* never happens */
      break;
    }

    if (next && (l == 0 || send(fd, buf, l, 0) >= 0)) {
      if (next == DO_CONN) {
	++c->state;
        event_request_io(ec, fd, EVENT_READ, connh, c);
      }
      else if (c->dest->type == DT_EXEC)
	execprog(ec, fd, c, next == DO_DEST ? 1 : 0);
      else {
        if (c->dest->type == DT_TEXPECT) {
          l = sprintf(buf, "%s:%s:%d\r\n", c->port->proto->name,
                      inet_ntoa(c->host->addr), c->port->port);
          write(fd, buf, l);
        }
        c->state = 0;
        event_request_io(ec, fd, EVENT_READ, expecth, c);
      }
      return;
    }
  }
  if (c->state < 0)
    cinfo(2, c, "connect: %s", errno ? strerror(errno) : "end of file");
  else
    cinfo(2, c, "%s", errno ? strerror(errno) : "EOF");
  connfree(ec, fd, c, 0);
}
